<?php
declare(strict_types=1);

namespace OCA\VibeEToEe\AppInfo;

use OCA\VibeEToEe\Capabilities;
use OCA\VibeEToEe\Connector\Sabre\LockPlugin;
use OCA\VibeEToEe\Connector\Sabre\RedirectRequestPlugin;
use OCA\VibeEToEe\Contracts\IExportVibe;
use OCA\VibeEToEe\Contracts\IRestApi;
use OCA\VibeEToEe\Db\UseradminMapper;
use OCA\VibeEToEe\EncryptionManager;
use OCA\VibeEToEe\Contracts\IKeyStorage;
use OCA\VibeEToEe\Contracts\IMetaDataStorage;
use OCA\VibeEToEe\Contracts\IStorageIdentify;
use OCA\VibeEToEe\Events\UnlinkedGroupsEvent;
use OCA\VibeEToEe\KeyStorage;
use OCA\VibeEToEe\Listener\BeforeUserCreatedListener;
use OCA\VibeEToEe\Listener\GroupDeletedListener;
use OCA\VibeEToEe\Listener\UnlinkedGroupsListener;
use OCA\VibeEToEe\Listener\UserAddedToGroupListener;
use OCA\VibeEToEe\Listener\UserChangedListener;
use OCA\VibeEToEe\Listener\UserCreatedListener;
use OCA\VibeEToEe\Listener\UserCreateListener;
use OCA\VibeEToEe\Listener\UserDeletedListener;
use OCA\VibeEToEe\Listener\UserRemovedFromGroupListener;
use OCA\VibeEToEe\MetaDataStorage;
use OCA\VibeEToEe\Middleware\UserAgentCheckMiddleware;
use OCA\Files_Trashbin\Events\MoveToTrashEvent;
use OCA\Files_Versions\Events\CreateVersionEvent;
use OCA\VibeEToEe\Service\ExportVibe;
use OCA\VibeEToEe\Service\RestApi;
use OCA\VibeEToEe\StorageIdentify;
use OCP\AppFramework\App;
use OCP\AppFramework\Bootstrap\IBootContext;
use OCP\AppFramework\Bootstrap\IBootstrap;
use OCP\AppFramework\Bootstrap\IRegistrationContext;
use OCP\EventDispatcher\IEventDispatcher;
use OCP\Group\Events\GroupDeletedEvent;
use OCP\Group\Events\UserAddedEvent;
use OCP\Group\Events\UserRemovedEvent;
use OCP\IGroup;
use OCP\SabrePluginEvent;
use OCP\User\Events\BeforeUserCreatedEvent;
use OCP\User\Events\CreateUserEvent;
use OCP\User\Events\UserChangedEvent;
use OCP\User\Events\UserCreatedEvent;
use OCP\User\Events\UserDeletedEvent;
use OCP\IUserSession;

use OC;
use OC\User\User;
use Psr\Container\ContainerInterface;


class Application extends App implements IBootstrap {

	public const APP_ID = 'vibeetoee';
    public const APP_NAME = 'VIBE E2E Encryption';
    /**
     * @var OC\AllConfig
     */
    protected $config;

    /**
     * @var User
     */
    protected $user;

    public $configVibe;


    /**
     * Application constructor.
     * @param array $urlParams
     * @throws \OCP\AppFramework\QueryException
     */
	public function __construct(array $urlParams = []) {
		parent::__construct(self::APP_ID, $urlParams);

		$container = $this->getContainer();
		$dispatcher = $container->query(IEventDispatcher::class);

        $this->config = OC::$server->getConfig();
        $this->user = OC::$server[IUserSession::class]->getUser();
        $this->configVibe = self::getConfigVibe();

//        $container->registerService('UseradminMapper', function(ContainerInterface $c){
//            return new AuthorService(
//                $c->get('AuthorMapper')
//            );
//        });
//        $container->registerService('UseradminMapper', function(ContainerInterface $c){
//            return new UseradminMapper(
//                $c->get('ServerContainer')->getDatabaseConnection()
//            );
//        });

        $this->registerEventsScripts($dispatcher);
	}


    /**
     * Checks if this app is enabled.
     */
    public function isEnabled(): bool
    {
        $enabled = true;
        $isForced = (bool) $this->config->getAppValue(self::APP_ID, 'force', '0');

        if (null !== $this->user && !$isForced) {
            $enabled = (bool) $this->config->getUserValue(
                $this->user->getUid(),
                self::APP_ID,
                'enabled',
                $this->config->getAppValue(
                    self::APP_ID,
                    'default-enabled',
                    '1'
                )
            );
        }

        return $enabled;
    }


    /**
     * @param IRegistrationContext $context
     */
	public function register(IRegistrationContext $context): void {
		$context->registerCapability(Capabilities::class);
		$context->registerMiddleware(UserAgentCheckMiddleware::class);

		$context->registerServiceAlias(IKeyStorage::class, KeyStorage::class);
		$context->registerServiceAlias(IMetaDataStorage::class, MetaDataStorage::class);
		$context->registerServiceAlias(IStorageIdentify::class, StorageIdentify::class);
		$context->registerServiceAlias(IRestApi::class, RestApi::class);
		$context->registerServiceAlias(IExportVibe::class, ExportVibe::class);

		$context->registerEventListener(UserDeletedEvent::class, UserDeletedListener::class);
		$context->registerEventListener(UserChangedEvent::class, UserChangedListener::class);
//		$context->registerEventListener(IGroup::class . '::postAddUser', UserAddedToGroupListener::class);
	}

    /**
     * @param IBootContext $context
     */
	public function boot(IBootContext $context): void {
		$eventDispatcher = $context->getServerContainer()->get(IEventDispatcher::class);
		$eventDispatcher->addListener('OCA\DAV\Connector\Sabre::addPlugin', function (SabrePluginEvent $event) {
			$server = $event->getServer();

			if ($server !== null) {
				$server->addPlugin($this->getContainer()->get(LockPlugin::class));
				$server->addPlugin($this->getContainer()->get(RedirectRequestPlugin::class));
			}
		});

		$eventDispatcher->addListener('OCA\Files_Trashbin::moveToTrash', function (MoveToTrashEvent $event) {
			$encryptionManager = $this->getContainer()->get(EncryptionManager::class);

			$node = $event->getNode();
			if ($encryptionManager->isEncryptedFile($node)) {
				$event->disableTrashBin();
			}
		});

		$eventDispatcher->addListener('OCA\Files_Versions::createVersion', function (CreateVersionEvent $event) {
			$encryptionManager = $this->getContainer()->get(EncryptionManager::class);

			$node = $event->getNode();
			if ($encryptionManager->isEncryptedFile($node)) {
				$event->disableVersions();
			}
		});

//        $eventDispatcher->addListener('OCP\IGroup::postAddUser', function (GenericEvent $event) {
//            var_dump($event);
//            die();
//        });

//        $eventDispatcher->addListener('OCP\IGroup::preDelete', function() {
//            throw new \Exception('This group has a connection with the VibeGroup');
//        });


    }

	protected function registerEventsScripts(IEventDispatcher $dispatcher) {
	    $dispatcher->addServiceListener(UserCreatedEvent::class, UserCreatedListener::class);
//	    $dispatcher->addServiceListener(UserDeletedEvent::class, UserDeletedListener::class);
//	    $dispatcher->addServiceListener(CreateUserEvent::class, UserCreateListener::class);
//	    $dispatcher->addServiceListener(BeforeUserCreatedEvent::class, BeforeUserCreatedListener::class);

	    $dispatcher->addServiceListener(UserAddedEvent::class, UserAddedToGroupListener::class);
	    $dispatcher->addServiceListener(UserRemovedEvent::class, UserRemovedFromGroupListener::class);
	    $dispatcher->addServiceListener(GroupDeletedEvent::class, GroupDeletedListener::class);
	    $dispatcher->addServiceListener(UnlinkedGroupsEvent::class, UnlinkedGroupsListener::class);
    }

    /**
     * @return array|mixed
     */
    static public function getConfigVibe(){
        try {
            $cfg = file_get_contents(__DIR__ . '/../../config.json');
            $cfg = json_decode($cfg);
            return $cfg;

        } catch (\Exception $e){
            return [];
        }
    }

}
