<?php

declare(strict_types=1);
namespace OCA\VibeEToEe\Connector\Sabre;

use OC\AppFramework\Http;
use OCA\DAV\Connector\Sabre\Directory;
use OCA\DAV\Connector\Sabre\Exception\Forbidden;
use OCA\DAV\Connector\Sabre\File;
use OCP\Files\FileInfo;
use OCP\Files\IRootFolder;
use OCP\Files\Node;
use OCP\IUserSession;
use Sabre\DAV\Exception\Conflict;
use Sabre\DAV\Exception\NotFound;
use Sabre\DAV\INode;
use Sabre\DAV\Server;
use Sabre\DAV\ServerPlugin;
use Exception;

abstract class APlugin extends ServerPlugin {

	protected $server;
	protected $rootFolder;
	protected $userSession;
	private $applyPlugin;

    /**
     * APlugin constructor.
     * @param IRootFolder $rootFolder
     * @param IUserSession $userSession
     */
	public function __construct(IRootFolder $rootFolder,
								IUserSession $userSession) {
		$this->rootFolder = $rootFolder;
		$this->userSession = $userSession;

		$this->applyPlugin = [];
	}

    /**
     * @param Server $server
     */
	public function initialize(Server $server) {
		$this->server = $server;
	}

    /**
     * @param string $path
     * @param string $method
     * @return INode
     * @throws Conflict
     */
	protected function getNode(string $path, string $method): INode {
		return $this->getNodeForPath($path);
	}

    /**
     * @param string $path
     * @return INode
     * @throws Conflict
     */
	protected function getNodeForPath(string $path): INode {
		if ($this->server->tree->nodeExists($path)) {
			return $this->server->tree->getNodeForPath($path);
		}

		$parent = dirname($path);
		$parent = ($parent === '.') ? '/' : $parent;
		if ($this->server->tree->nodeExists($parent)) {
			return $this->server->tree->getNodeForPath($parent);
		}

		throw new Conflict();
	}

    /**
     * @param string $path
     * @return Node
     * @throws Forbidden
     * @throws NotFound
     */
	protected function getFileNode(string $path): Node {
		$user = $this->userSession->getUser();
		if ($user === null) {
			throw new Forbidden('No user session found');
		}
		$uid = $user->getUID();

		try {
			return $this->rootFolder
				->getUserFolder($uid)
				->get($path);
		} catch (Exception $e) {
			throw new NotFound('file not found', Http::STATUS_NOT_FOUND, $e);
		}
	}

    /**
     * @param string $url
     * @param INode $node
     * @return bool
     */
	protected function isFile(string $url, INode $node): bool {
		if (isset($this->applyPlugin[$url])) {
			return $this->applyPlugin[$url];
		}

		$this->applyPlugin[$url] = (($node instanceof File) || ($node instanceof Directory));

		return $this->applyPlugin[$url];
	}

    /**
     * @param string $path
     * @return bool
     * @throws Forbidden
     */
	protected function isE2EEnabledPath(string $path):bool {
		try {
			$node = $this->getFileNode($path);
		} catch (NotFound $e) {
			return false;
		}

		while ($node->isEncrypted() === false || $node->getType() === FileInfo::TYPE_FILE) {
			$node = $node->getParent();

			if ($node->getPath() === '/') {
				return false;
			}
		}

		return true;
	}
}
