<?php

declare(strict_types=1);

namespace OCA\VibeEToEe\Connector\Sabre;

use OC\AppFramework\Http;
use OCA\DAV\Connector\Sabre\Directory;
use OCA\DAV\Connector\Sabre\Exception\FileLocked;
use OCA\DAV\Connector\Sabre\Exception\Forbidden;
use OCA\DAV\Connector\Sabre\File;
use OCA\DAV\Upload\FutureFile;
use OCA\VibeEToEe\LockManager;
use OCA\VibeEToEe\UserAgentManager;
use OCP\Files\IRootFolder;
use OCP\IUserSession;
use Sabre\DAV\Exception\Conflict;
use Sabre\DAV\Exception\NotFound;
use Sabre\DAV\INode;
use Sabre\DAV\Server;
use Sabre\HTTP\RequestInterface;

class LockPlugin extends APlugin {

	private $lockManager;
	private $userAgentManager;

    /**
     * LockPlugin constructor.
     * @param IRootFolder $rootFolder
     * @param IUserSession $userSession
     * @param LockManager $lockManager
     * @param UserAgentManager $userAgentManager
     */
	public function __construct(IRootFolder $rootFolder,
								IUserSession $userSession,
								LockManager $lockManager,
								UserAgentManager $userAgentManager) {
		parent::__construct($rootFolder, $userSession);
		$this->lockManager = $lockManager;
		$this->userAgentManager = $userAgentManager;
	}

    /**
     * @param Server $server
     */
	public function initialize(Server $server) {
		parent::initialize($server);

		$this->server->on('beforeMethod:DELETE', [$this, 'checkLock'], 200);
		$this->server->on('beforeMethod:MKCOL', [$this, 'checkLock'], 200);
		$this->server->on('beforeMethod:PUT', [$this, 'checkLock'], 200);

		$this->server->on('beforeMethod:COPY', [$this, 'checkLock'], 200);
		$this->server->on('beforeMethod:MOVE', [$this, 'checkLock'], 200);
	}

    /**
     * @param RequestInterface $request
     * @throws Conflict
     * @throws FileLocked
     * @throws Forbidden
     */
	public function checkLock(RequestInterface $request): void {
		$node = $this->getNode($request->getPath(), $request->getMethod());
		$url = $request->getAbsoluteUrl();
		$method = $request->getMethod();

		if (!$this->isFile($url, $node)) {
			return;
		}
		/** @var File|Directory|FutureFile $node*/

		if ($method === 'COPY' || $method === 'MOVE') {
			$destInfo = $this->server->getCopyAndMoveInfo($request);
			$destNode = $this->getNode($destInfo['destination'], $method);

			if ($node instanceof FutureFile) {
				if ($this->isE2EEnabledPath($destNode->getPath()) === false) {
					return;
				}
			} else {
				if (!$this->isE2EEnabledPath($node->getPath()) && !$this->isE2EEnabledPath($destNode->getPath())) {
					return;
				}

				if ($this->isE2EEnabledPath($node->getPath()) xor $this->isE2EEnabledPath($destNode->getPath())) {
					throw new Forbidden('Cannot copy or move files from non-encrypted folders to end to end encrypted folders or vice versa.');
				}
			}
		} elseif (!$this->isE2EEnabledPath($node->getPath())) {
			return;
		}

		$userAgent = $request->getHeader('user-agent');
		if (!$this->isE2EEnabledUserAgent($userAgent)) {
			throw new Forbidden('Client "' . $userAgent . '" is not allowed to access end-to-end encrypted content');
		}

		$e2eToken = null;
		if ($request->hasHeader('e2e-token')) {
			$e2eToken = $request->getHeader('e2e-token');
		} else {
			$queryParams = $request->getQueryParameters();
			if (array_key_exists('e2e-token', $queryParams)) {
				$e2eToken = $queryParams['e2e-token'];
			}
		}

		switch ($method) {
			case 'COPY':
			case 'MOVE':
				$node instanceof FutureFile || $this->verifyTokenOnWriteAccess($node, $e2eToken);
				$this->verifyTokenOnWriteAccess($destNode, $e2eToken);
				break;

			default:
				$this->verifyTokenOnWriteAccess($node, $e2eToken);
				break;
		}
	}

    /**
     * @param INode $node
     * @param string|null $token
     * @throws FileLocked
     * @throws Forbidden
     */
	protected function verifyTokenOnWriteAccess(INode $node, ?string $token): void {
		if ($token === null) {
			throw new Forbidden('Write access to end-to-end encrypted folder requires token - no token sent');
		}

		if ($this->lockManager->isLocked($node->getId(), $token)) {
			throw new FileLocked('Write access to end-to-end encrypted folder requires token - resource not locked or wrong token sent', Http::STATUS_FORBIDDEN);
		}
	}

    /**
     * @param string $userAgent
     * @return bool
     */
	protected function isE2EEnabledUserAgent(string $userAgent):bool {
		return $this->userAgentManager->supportsVibeEToEe($userAgent);
	}
}
