<?php

declare(strict_types=1);

namespace OCA\VibeEToEe\Controller;

use OCA\VibeEToEe\Exceptions\KeyExistsException;
use OCA\VibeEToEe\Contracts\IKeyStorage;
use OCA\VibeEToEe\SignatureHandler;
use OCP\AppFramework\Http;
use OCP\AppFramework\Http\DataResponse;
use OCP\AppFramework\OCS\OCSBadRequestException;
use OCP\AppFramework\OCS\OCSForbiddenException;
use OCP\AppFramework\OCS\OCSNotFoundException;
use OCP\AppFramework\OCSController;
use OCP\Files\ForbiddenException;
use OCP\Files\NotFoundException;
use OCP\Files\NotPermittedException;
use OCP\IL10N;
use OCP\ILogger;
use OCP\IRequest;
use \Exception;
use \BadMethodCallException;

class KeyController extends OCSController {

	private $userId;
	private $keyStorage;
	private $signatureHandler;
	private $logger;
	private $l10n;

    /**
     * KeyController constructor.
     * @param $AppName
     * @param IRequest $request
     * @param $userId
     * @param IKeyStorage $keyStorage
     * @param SignatureHandler $signatureHandler
     * @param ILogger $logger
     * @param IL10N $l10n
     */
	public function __construct($AppName,
								IRequest $request,
								$userId,
								IKeyStorage $keyStorage,
								SignatureHandler $signatureHandler,
								ILogger $logger,
								IL10N $l10n
	) {
		parent::__construct($AppName, $request);
		$this->userId = $userId;
		$this->keyStorage = $keyStorage;
		$this->signatureHandler = $signatureHandler;
		$this->logger = $logger;
		$this->l10n = $l10n;
	}

    /**
     * @return DataResponse
     * @throws OCSBadRequestException
     * @throws OCSForbiddenException
     * @throws OCSNotFoundException
     *
     * @NoAdminRequired
     */
	public function getPrivateKey(): DataResponse {
		try {
			$privateKey = $this->keyStorage->getPrivateKey($this->userId);
			return new DataResponse(['private-key' => $privateKey]);
		} catch (ForbiddenException $e) {
			throw new OCSForbiddenException($this->l10n->t('This is someone else\'s private key'));
		} catch (NotFoundException $e) {
			throw new OCSNotFoundException($this->l10n->t('Could not find the private key of the user %s', [$this->userId]));
		} catch (Exception $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($this->l10n->t('Internal error'));
		}
	}

    /**
     * @return DataResponse
     * @throws OCSBadRequestException
     * @throws OCSForbiddenException
     * @throws OCSNotFoundException
     *
     * @NoAdminRequired
     */
	public function deletePrivateKey(): DataResponse {
		try {
			$this->keyStorage->deletePrivateKey($this->userId);
			return new DataResponse();
		} catch (NotPermittedException $e) {
			throw new OCSForbiddenException($this->l10n->t('You are not allowed to delete this private key'));
		} catch (NotFoundException $e) {
			throw new OCSNotFoundException($this->l10n->t('Could not find the private key belonging to the user %s', [$this->userId]));
		} catch (Exception $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($this->l10n->t('Internal error'));
		}
	}


    /**
     * @param string $privateKey
     * @return DataResponse
     * @throws OCSBadRequestException
     *
     * @NoAdminRequired
     */
	public function setPrivateKey(string $privateKey): DataResponse {
		try {
			$this->keyStorage->setPrivateKey($privateKey, $this->userId);
		} catch (KeyExistsException $e) {
            $this->logger->logException($e, ['app' => $this->appName]);
			return new DataResponse([], Http::STATUS_CONFLICT);
		} catch (Exception $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($this->l10n->t('Internal error'));
		}

		return new DataResponse(['private-key' => $privateKey]);
	}

    /**
     * @param string $users
     * @return DataResponse
     * @throws OCSBadRequestException
     * @throws OCSNotFoundException
     *
     * @NoAdminRequired
     */
	public function getPublicKeys(string $users = ''): DataResponse {
		$usersArray = $this->jsonDecode($users);

		$result = ['public-keys' => []];
		foreach ($usersArray as $uid) {
			try {
				$publicKey = $this->keyStorage->getPublicKey($uid);
				$result['public-keys'][$uid] = $publicKey;
			} catch (NotFoundException $e) {
				throw new OCSNotFoundException($this->l10n->t('Could not find the public key belonging to the user %s', [$uid]));
			} catch (Exception $e) {
				$this->logger->logException($e, ['app' => $this->appName]);
				throw new OCSBadRequestException($this->l10n->t('Internal error'));
			}
		}

		return new DataResponse($result);
	}

    /**
     * @param string $csr
     * @return DataResponse
     * @throws OCSBadRequestException
     * @throws OCSForbiddenException
     *
     * @NoAdminRequired
     */
	public function createPublicKey(string $csr): DataResponse {
		if ($this->keyStorage->publicKeyExists($this->userId)) {
			return new DataResponse([], Http::STATUS_CONFLICT);
		}

		try {
			$subject = openssl_csr_get_subject($csr);
			$publicKey = $this->signatureHandler->sign($csr);
		} catch (BadMethodCallException $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($e->getMessage());
		} catch (Exception $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($this->l10n->t('Internal error'));
		}

		$cn = isset($subject['CN']) ? $subject['CN'] : '';
		if ($cn !== $this->userId) {
			throw new OCSForbiddenException($this->l10n->t('Common name (CN) does not match the current user'));
		}

		$this->keyStorage->setPublicKey($publicKey, $this->userId);

		return new DataResponse(['public-key' => $publicKey]);
	}

    /**
     * @return DataResponse|null
     * @throws OCSBadRequestException
     * @throws OCSForbiddenException
     * @throws OCSNotFoundException
     *
     * @NoAdminRequired
     */
	public function deletePublicKey(): ?DataResponse {
		try {
			$this->keyStorage->deletePublicKey($this->userId);
			return new DataResponse();
		} catch (NotFoundException $e) {
			throw new OCSNotFoundException($this->l10n->t('Could not find the public key belonging to %s', [$this->userId]));
		} catch (NotPermittedException $e) {
			throw new OCSForbiddenException($this->l10n->t('This is not your public key to delete'));
		} catch (Exception $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($this->l10n->t('Internal error'));
		}
	}


    /**
     * @return DataResponse
     * @throws OCSBadRequestException
     *
     * @NoAdminRequired
     */
	public function getPublicServerKey(): DataResponse {
		try {
			$publicKey = $this->signatureHandler->getPublicServerKey();
		} catch (Exception $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($this->l10n->t('Internal error'));
		}

		return new DataResponse(['public-key' => $publicKey]);
	}

    /**
     * @param string $users
     * @return array
     * @throws OCSBadRequestException
     */
	private function jsonDecode(string $users): array {
		$usersArray = [];
		if (!empty($users)) {
			$usersArray = \json_decode($users, true);
			if ($usersArray === null) {
				throw new OCSBadRequestException($this->l10n->t('Can not decode userlist'));
			}
		}

		if (!in_array($this->userId, $usersArray, true)) {
			$usersArray[] = $this->userId;
		}

		return $usersArray;
	}
}
