<?php

declare(strict_types=1);

namespace OCA\VibeEToEe\Controller;

use OCA\VibeEToEe\AppInfo\Application;
use OC\User\NoUserException;
use OCA\VibeEToEe\Exceptions\FileLockedException;
use OCA\VibeEToEe\Exceptions\FileNotLockedException;
use OCA\VibeEToEe\FileService;
use OCA\VibeEToEe\Contracts\IMetaDataStorage;
use OCA\VibeEToEe\LockManager;
use OCP\AppFramework\Http\DataResponse;
use OCP\AppFramework\OCS\OCSForbiddenException;
use OCP\AppFramework\OCS\OCSNotFoundException;
use OCP\AppFramework\OCSController;
use OCP\Files\Folder;
use OCP\Files\IRootFolder;
use OCP\IL10N;
use OCP\IRequest;
use OCP\IConfig;

class LockingController extends OCSController {

    /** @var IConfig */
    protected $config;
    private $userId;
    private $metaDataStorage;
    private $rootFolder;
    private $fileService;
    private $lockManager;
    private $l10n;

    public function __construct($AppName,
                                IRequest $request,
                                $userId,
                                IMetaDataStorage $metaDataStorage,
                                LockManager $lockManager,
                                IRootFolder $rootFolder,
                                FileService $fileService,
                                IConfig $config,
                                IL10N $l10n
    ) {
        parent::__construct($AppName, $request);
        $this->userId = $userId;
        $this->metaDataStorage = $metaDataStorage;
        $this->rootFolder = $rootFolder;
        $this->fileService = $fileService;
        $this->lockManager = $lockManager;
        $this->l10n = $l10n;
        $this->config = $config;
    }

    /**
     * @param int $id
     * @return DataResponse
     * @throws OCSForbiddenException
     *
     * @NoAdminRequired
     */
    public function lockFolder(int $id): DataResponse {
        $e2eToken = $this->request->getParam('e2e-token', '');
        $newToken = $this->lockManager->lockFile($id, $e2eToken);

        if ($newToken === null) {

            if (((time() - $this->lockManager->getOldTimestamp()) / 60) > $this->config->getAppValue(Application::APP_ID, 'vibe-timeout', 120)) {
                // ---------------------------
                try {
                    //$this->unlockProcess($id, $this->lockManager->getOldToken());
                    $this->lockManager->unlockFile($id, $this->lockManager->getOldToken());

                } catch (NoUserException $e) {
                    throw new OCSForbiddenException($this->l10n->t('You are not allowed to remove the lock'));
                } catch (FileLockedException $e) {
                    throw new OCSForbiddenException($this->l10n->t('You are not allowed to remove the lock'));
                } catch (FileNotLockedException $e) {
                    throw new OCSNotFoundException($this->l10n->t('File not locked'));
                } catch (\Exception $e) {

                }

                $newToken = $this->lockManager->lockFile($id, $e2eToken);
                if ($newToken === null)
                    throw new OCSForbiddenException($this->l10n->t('File CRITICAL locked'));
                // ---------------------------

            } else {
                throw new OCSForbiddenException($this->l10n->t('File already locked'));
            }
        }

        return new DataResponse(['e2e-token' => $newToken]);
    }


    /**
     * @param int $id
     * @return DataResponse
     * @throws OCSForbiddenException
     * @throws OCSNotFoundException
     * @throws \OCP\Files\NotPermittedException
     *
     * @NoAdminRequired
     */
    public function unlockFolder(int $id): DataResponse {
        $token = $this->request->getHeader('e2e-token');


        try {
            $this->unlockProcess($id, $token);

        } catch (NoUserException $e) {
            throw new OCSForbiddenException($this->l10n->t('You are not allowed to remove the lock'));
        } catch (FileLockedException $e) {
            throw new OCSForbiddenException($this->l10n->t('You are not allowed to remove the lock'));
        } catch (FileNotLockedException $e) {
            throw new OCSNotFoundException($this->l10n->t('File not locked'));
        } catch (\Exception $e){

        }

        return new DataResponse();
    }

    protected function unlockProcess($id = null, $token = null){

        try {
            $userFolder = $this->rootFolder->getUserFolder($this->userId);
        } catch (NoUserException $e) {
            throw $e;
//            throw new OCSForbiddenException($this->l10n->t('You are not allowed to remove the lock'));
        }

        $nodes = $userFolder->getById($id);

        if (!isset($nodes[0]) || !$nodes[0] instanceof Folder) {
            throw new OCSForbiddenException($this->l10n->t('You are not allowed to remove the lock'));
        }

        $this->fileService->finalizeChanges($nodes[0]);

        try {
            $this->metaDataStorage->saveIntermediateFile($this->userId, $id);
        } catch (\Exception $e){
            throw  $e;
        }

        try {
            $this->lockManager->unlockFile($id, $token);
        } catch (FileLockedException $e) {
            throw $e;
//            throw new OCSForbiddenException($this->l10n->t('You are not allowed to remove the lock'));
        } catch (FileNotLockedException $e) {
            throw $e;
//            throw new OCSNotFoundException($this->l10n->t('File not locked'));
        }

    }
}
