<?php

declare(strict_types=1);

namespace OCA\VibeEToEe\Controller;

use OCA\VibeEToEe\Exceptions\MetaDataExistsException;
use OCA\VibeEToEe\Exceptions\MissingMetaDataException;
use OCA\VibeEToEe\Contracts\IMetaDataStorage;
use OCA\VibeEToEe\LockManager;
use OCP\AppFramework\Http;
use OCP\AppFramework\Http\DataResponse;
use OCP\AppFramework\OCS\OCSBadRequestException;
use OCP\AppFramework\OCS\OCSForbiddenException;
use OCP\AppFramework\OCS\OCSNotFoundException;
use OCP\AppFramework\OCSController;
use OCP\Files\NotFoundException;
use OCP\Files\NotPermittedException;
use OCP\IL10N;
use OCP\ILogger;
use OCP\IRequest;

class MetaDataController extends OCSController {

	private $userId;
	private $metaDataStorage;
	private $logger;
	private $lockManager;
	private $l10n;

	public function __construct($AppName,
								IRequest $request,
								$userId,
								IMetaDataStorage $metaDataStorage,
								LockManager $lockManager,
								ILogger $logger,
								IL10N $l10n
	) {
		parent::__construct($AppName, $request);
		$this->userId = $userId;
		$this->metaDataStorage = $metaDataStorage;
		$this->logger = $logger;
		$this->lockManager = $lockManager;
		$this->l10n = $l10n;
	}

    /**
     * @param int $id
     * @return DataResponse
     * @throws OCSBadRequestException
     * @throws OCSNotFoundException
     *
     * @NoAdminRequired
     */
	public function getMetaData(int $id): DataResponse {
		try {
			$metaData = $this->metaDataStorage->getMetaData($this->userId, $id);
		} catch (NotFoundException $e) {
			throw new OCSNotFoundException($this->l10n->t('Could not find metadata for "%s"', [$id]));
		} catch (\Exception $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($this->l10n->t('Can\'t read metadata'));
		}
		return new DataResponse(['meta-data' => $metaData]);
	}

    /**
     * @param int $id
     * @param string $metaData
     * @return DataResponse
     * @throws OCSBadRequestException
     * @throws OCSNotFoundException
     *
     * @NoAdminRequired
     */
	public function setMetaData(int $id, string $metaData): DataResponse {
		try {
			$this->metaDataStorage->setMetaDataIntoIntermediateFile($this->userId, $id, $metaData);
		} catch (MetaDataExistsException $e) {
			return new DataResponse([], Http::STATUS_CONFLICT);
		} catch (NotFoundException $e) {
			throw new OCSNotFoundException($e->getMessage());
		} catch (\Exception $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($this->l10n->t('Can\'t store metadata'));
		}

		return new DataResponse(['meta-data' => $metaData]);
	}

    /**
     * @param int $id
     * @param string $metaData
     * @return DataResponse
     * @throws OCSBadRequestException
     * @throws OCSForbiddenException
     * @throws OCSNotFoundException
     *
     * @NoAdminRequired
     */
	public function updateMetaData(int $id, string $metaData): DataResponse {
		$e2eToken = $this->request->getParam('e2e-token');

		if ($this->lockManager->isLocked($id, $e2eToken)) {
			throw new OCSForbiddenException($this->l10n->t('You are not allowed to edit the file, make sure to first lock it, and then send the right token'));
		}

		try {
			$this->metaDataStorage->updateMetaDataIntoIntermediateFile($this->userId, $id, $metaData);
		} catch (MissingMetaDataException $e) {
			throw new OCSNotFoundException($this->l10n->t('Metadata-file doesn\'t exist'));
		} catch (NotFoundException $e) {
			throw new OCSNotFoundException($e->getMessage());
		} catch (\Exception $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($this->l10n->t('Can\'t store metadata'));
		}

		return new DataResponse(['meta-data' => $metaData]);
	}

    /**
     * @param int $id
     * @return DataResponse
     * @throws OCSBadRequestException
     * @throws OCSForbiddenException
     * @throws OCSNotFoundException
     *
     * @NoAdminRequired
     */
	public function deleteMetaData(int $id): DataResponse {
		try {
			$this->metaDataStorage->updateMetaDataIntoIntermediateFile($this->userId, $id, '{}');
		} catch (NotFoundException $e) {
			throw new OCSNotFoundException($this->l10n->t('Could not find metadata for "%s"', [$id]));
		} catch (NotPermittedException $e) {
			throw new OCSForbiddenException($this->l10n->t('Only the owner can delete the metadata-file'));
		} catch (\Exception $e) {
			$this->logger->logException($e, ['app' => $this->appName]);
			throw new OCSBadRequestException($this->l10n->t('Can\'t delete metadata'));
		}
		return new DataResponse();
	}
}
