<?php

declare(strict_types=1);

namespace OCA\VibeEToEe;

use OCP\Files\Folder;
use OCP\Files\IRootFolder;
use OCP\Files\Node;
use OCP\Files\NotFoundException;
use OCP\IUserSession;
use OCP\Share\IManager;
use OCP\Share\IShare;

class EncryptionManager {

	private $rootFolder;
	private $userSession;
	private $shareManager;

	public function __construct(IRootFolder $rootFolder,
								IUserSession $userSession,
								IManager $shareManager) {
		$this->rootFolder = $rootFolder;
		$this->userSession = $userSession;
		$this->shareManager = $shareManager;
	}

    /**
     * @param int $id
     * @throws NotFoundException
     * @throws \OCP\Files\NotPermittedException
     * @throws \OC\User\NoUserException
     */
	public function setEncryptionFlag(int $id): void {
		$this->isValidFolder($id);
		$userRoot = $this->getUserRoot();
		$userRoot->getStorage()->getCache()->update($id, ['encrypted' => '1']);
	}

    /**
     * @param int $id
     * @throws NotFoundException
     * @throws \OCP\Files\NotPermittedException
     * @throws \OC\User\NoUserException
     */
	public function removeEncryptionFlag(int $id): void {
		$this->isValidFolder($id);
		$userRoot = $this->getUserRoot();
		$userRoot->getStorage()->getCache()->update($id, ['encrypted' => '0']);
	}

    /**
     * @param Node $node
     * @return bool
     */
	public function isEncryptedFile(Node $node): bool {
		do {
			if ($node->isEncrypted()) {
				return true;
			}
			$node = $node->getParent();
		} while ($node->getPath() !== '/');

		return false;
	}

    /**
     * @return Folder
     * @throws \OCP\Files\NotPermittedException
     * @throws \OC\User\NoUserException
     */
	protected function getUserRoot(): Folder {
		$uid = $this->userSession->getUser()->getUID();
		$userRoot = $this->rootFolder->getUserFolder($uid);
		return $userRoot;
	}


    /**
     * @param int $id
     * @throws NotFoundException
     */
	protected function isValidFolder(int $id):void {
		$node = $this->rootFolder->getById($id);

		if (!isset($node[0])) {
			throw new NotFoundException('No folder with ID ' . $id);
		}

		$firstNode = $node[0];
		if (!($firstNode instanceof Folder)) {
			throw new NotFoundException('No folder with ID ' . $id);
		}

		if (!empty($firstNode->getDirectoryListing())) {
			throw new NotFoundException('Folder with ID ' . $id . ' not empty');
		}

		$user = $this->userSession->getUser();
		if ($user === null) {
			throw new NotFoundException('No active user-session');
		}

		$userId = $user->getUID();
		if ($this->isNodeShared($userId, $firstNode)) {
			throw new NotFoundException('Folder with ID ' . $id . ' is shared');
		}
	}

    /**
     * @param string $userId
     * @param Node $node
     * @return bool
     */
	private function isNodeShared(string $userId, Node $node):bool {
		$shareTypesToCheck = [
			IShare::TYPE_USER,
			IShare::TYPE_GROUP,
			IShare::TYPE_USERGROUP,
			IShare::TYPE_LINK,
			IShare::TYPE_EMAIL,
			IShare::TYPE_REMOTE,
			IShare::TYPE_CIRCLE,
			IShare::TYPE_GUEST,
			IShare::TYPE_REMOTE_GROUP,
			IShare::TYPE_ROOM,
		];

		foreach ($shareTypesToCheck as $shareType) {
			$shares = $this->shareManager->getSharesBy(
				$userId,
				$shareType,
				$node,
				false,
				1 // Limit 1, because we only care whether there is a share or not
			);

			if (!empty($shares)) {
				return true;
			}
		}

		return false;
	}
}
