<?php

declare(strict_types=1);
namespace OCA\VibeEToEe;

use OCA\VibeEToEe\Connector\Sabre\RedirectRequestPlugin;
use OCP\Files\Folder;
use OCP\Files\Node;
use OCP\ILogger;

class FileService {

	public function __construct(ILogger $logger) {
		$this->logger = $logger;
	}

    /**
     * @param Folder $folder
     * @return bool
     * @throws \OCP\Files\InvalidPathException
     * @throws \OCP\Files\NotFoundException
     * @throws \OCP\Files\NotPermittedException
     * @throws \OCP\Lock\LockedException
     */
	public function revertChanges(Folder $folder): bool {
		$intermediateFiles = $this->getIntermediateFiles($folder);
		if (empty($intermediateFiles['to_save']) && empty($intermediateFiles['to_delete'])) {
			return false;
		}

		foreach ($intermediateFiles['to_save'] as $intermediateFile) {
			$intermediateFile->delete();
		}

		foreach ($intermediateFiles['to_delete'] as $intermediateFile) {
			$newPath = $this->removeSuffixFromString($intermediateFile->getPath(), RedirectRequestPlugin::DELETE_SUFFIX);
			$intermediateFile->move($newPath);
		}

		return true;
	}

    /**
     * @param Folder $folder
     * @return bool
     * @throws \OCP\Files\InvalidPathException
     * @throws \OCP\Files\NotFoundException
     * @throws \OCP\Files\NotPermittedException
     * @throws \OCP\Lock\LockedException
     */
	public function finalizeChanges(Folder $folder): bool {
		$intermediateFiles = $this->getIntermediateFiles($folder);
		if (empty($intermediateFiles['to_save']) && empty($intermediateFiles['to_delete'])) {
			return false;
		}

		/** @var Node $intermediateFile */
		foreach ($intermediateFiles['to_save'] as $intermediateFile) {
			$newPath = $this->removeSuffixFromString($intermediateFile->getPath(), RedirectRequestPlugin::SAVE_SUFFIX);
			$intermediateFile->move($newPath);
		}

		/** @var Node $intermediateFile */
		foreach ($intermediateFiles['to_delete'] as $intermediateFile) {
			$intermediateFile->delete();
		}

		return true;
	}

    /**
     * @param Folder $folder
     * @return array
     * @throws \OCP\Files\NotFoundException
     */
	private function getIntermediateFiles(Folder $folder): array {
		$listing = $folder->getDirectoryListing();
		$result = [
			'to_save' => [],
			'to_delete' => [],
		];

		foreach ($listing as $node) {
			if ($this->isIntermediateFileToSave($node)) {
				$result['to_save'][] = $node;
				continue;
			}

			if ($this->isIntermediateFileToDelete($node)) {
				$result['to_delete'][] = $node;
			}
		}

		return $result;
	}

    /**
     * @param Node $node
     * @return bool
     */
	private function isIntermediateFileToSave(Node $node): bool {
		return (substr($node->getName(), strlen(RedirectRequestPlugin::SAVE_SUFFIX) * -1) === RedirectRequestPlugin::SAVE_SUFFIX);
	}

    /**
     * @param Node $node
     * @return bool
     */
	private function isIntermediateFileToDelete(Node $node): bool {
		return (substr($node->getName(), strlen(RedirectRequestPlugin::DELETE_SUFFIX) * -1) === RedirectRequestPlugin::DELETE_SUFFIX);
	}

    /**
     * @param string $filename
     * @param string $suffix
     * @return string
     */
	private function removeSuffixFromString(string $filename, string $suffix):string {
		if (substr($filename, strlen($suffix) * -1) !== $suffix) {
			return $filename;
		}

		return substr($filename, 0, strlen($filename) - strlen($suffix));
	}
}
