<?php

declare(strict_types=1);

namespace OCA\VibeEToEe;

use OCA\VibeEToEe\Contracts\IKeyStorage;
use OCA\VibeEToEe\Exceptions\KeyExistsException;
use OCP\Files\ForbiddenException;
use OCP\Files\IAppData;
use OCP\Files\NotFoundException;
use OCP\Files\NotPermittedException;
use OCP\IUser;
use OCP\IUserSession;

class KeyStorage implements IKeyStorage {

	private $appData;
	private $userSession;
	private $privateKeysRoot = '/private-keys';
	private $publicKeysRoot = '/public-keys';

	public function __construct(IAppData $appData,
								IUserSession $userSession) {
		$this->appData = $appData;
		$this->userSession = $userSession;
	}

    /**
     * @param string $uid
     * @return string
     * @throws NotFoundException
     * @throws NotPermittedException
     */
	public function getPublicKey(string $uid): string {
		$this->verifyFolderStructure();

		$fileName = $this->getFileNameForPublicKey($uid);
		return $this->appData->getFolder($this->publicKeysRoot)
			->getFile($fileName)
			->getContent();
	}

    /**
     * @param string $uid
     * @return bool
     * @throws NotFoundException
     * @throws NotPermittedException
     */
	public function publicKeyExists(string $uid): bool {
		$this->verifyFolderStructure();

		$fileName = $this->getFileNameForPublicKey($uid);
		return $this->appData->getFolder($this->publicKeysRoot)
			->fileExists($fileName);
	}

    /**
     * @param string $publicKey
     * @param string $uid
     * @throws NotFoundException
     * @throws NotPermittedException
     */
	public function setPublicKey(string $publicKey, string $uid): void {
		$this->verifyFolderStructure();

		$fileName = $this->getFileNameForPublicKey($uid);
		$publicKeyRoot = $this->appData->getFolder($this->publicKeysRoot);

//		if ($publicKeyRoot->fileExists($fileName)) {
//			throw new KeyExistsException('Public key already exists');
//		}

		$publicKeyRoot
			->newFile($fileName)
			->putContent($publicKey);
	}

    /**
     * @param string $uid
     * @throws NotFoundException
     * @throws NotPermittedException
     */
	public function deletePublicKey(string $uid): void {
		$this->verifyFolderStructure();

		$user = $this->userSession->getUser();
		if ($user === null || $user->getUID() !== $uid) {
			throw new NotPermittedException('You are not allowed to delete the public key');
		}

		$fileName = $this->getFileNameForPublicKey($uid);
		$publicKeyRoot = $this->appData->getFolder($this->publicKeysRoot);
		try {
			$file = $publicKeyRoot->getFile($fileName);
		} catch (NotFoundException $ex) {
			return;
		}

		$file->delete();
	}

    /**
     * @param string $uid
     * @return string
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws NotPermittedException
     */
	public function getPrivateKey(string $uid): string {
		$this->verifyFolderStructure();

		$user = $this->userSession->getUser();
		if ($user === null || $user->getUID() !== $uid) {
			throw new ForbiddenException('You are not allowed to access the private key', false);
		}

		$fileName = $this->getFileNameForPrivateKey($uid);
		return $this->appData->getFolder($this->privateKeysRoot)
			->getFile($fileName)
			->getContent();
	}

    /**
     * @param string $uid
     * @return bool
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws NotPermittedException
     */
	public function privateKeyExists(string $uid): bool {
		$this->verifyFolderStructure();

		$user = $this->userSession->getUser();
		if ($user === null || $user->getUID() !== $uid) {
			throw new ForbiddenException('You are not allowed to access the private key', false);
		}

		$fileName = $this->getFileNameForPrivateKey($uid);
		return $this->appData->getFolder($this->privateKeysRoot)
			->fileExists($fileName);
	}

    /**
     * @param string $privateKey
     * @param string $uid
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws NotPermittedException
     */
	public function setPrivateKey(string $privateKey, string $uid): void {
		$this->verifyFolderStructure();

		$user = $this->userSession->getUser();
		if ($user === null || $user->getUID() !== $uid) {
			throw new ForbiddenException('You are not allowed to write the private key', false);
		}

		$fileName = $this->getFileNameForPrivateKey($uid);
		$privateKeysRoot = $this->appData->getFolder($this->privateKeysRoot);

//		if ($privateKeysRoot->fileExists($fileName)) {
//			throw new KeyExistsException('Private key already exists');
//		}
//

        $privateKeysRoot
                ->newFile($fileName)
                ->putContent($privateKey);

	}

    /**
     * @param string $uid
     * @throws NotFoundException
     * @throws NotPermittedException
     */
	public function deletePrivateKey(string $uid): void {
		$this->verifyFolderStructure();

		$user = $this->userSession->getUser();
		if ($user === null || $user->getUID() !== $uid) {
			throw new NotPermittedException('You are not allowed to delete the private key');
		}

		$fileName = $this->getFileNameForPrivateKey($uid);
		$privateKeysRoot = $this->appData->getFolder($this->privateKeysRoot);
		try {
			$file = $privateKeysRoot->getFile($fileName);
		} catch (NotFoundException $ex) {
			return;
		}

		$file->delete();
	}

    /**
     * @param IUser $user
     * @throws NotFoundException
     * @throws NotPermittedException
     */
	public function deleteUserKeys(IUser $user): void {
		$this->verifyFolderStructure();
		$uid = $user->getUID();

		$this->deleteUsersPublicKey($uid);
		$this->deleteUsersPrivateKey($uid);
	}

    /**
     * @param string $uid
     * @throws NotPermittedException
     */
	protected function deleteUsersPublicKey(string $uid): void {
		$fileName = $this->getFileNameForPublicKey($uid);

		try {
			$this->appData->getFolder($this->publicKeysRoot)
				->getFile($fileName)
				->delete();
		} catch (NotFoundException $e) {
			return;
		}
	}

    /**
     * @param string $uid
     * @throws NotPermittedException
     */
	protected function deleteUsersPrivateKey(string $uid): void {
		$fileName = $this->getFileNameForPrivateKey($uid);

		try {
			$this->appData->getFolder($this->privateKeysRoot)
				->getFile($fileName)
				->delete();
		} catch (NotFoundException $e) {
			return;
		}
	}

    /**
     * @throws NotFoundException
     * @throws NotPermittedException
     */
	protected function verifyFolderStructure(): void {
		$keyStorageRoot = $this->appData->getFolder('/');
		if (!$keyStorageRoot->fileExists($this->privateKeysRoot)) {
			$this->appData->newFolder($this->privateKeysRoot);
		}
		if (!$keyStorageRoot->fileExists($this->publicKeysRoot)) {
			$this->appData->newFolder($this->publicKeysRoot);
		}
	}

    /**
     * @param string $uid
     * @return string
     */
	private function getFileNameForPublicKey(string $uid):string {
		return $uid . '.public.key';
	}

    /**
     * @param string $uid
     * @return string
     */
	private function getFileNameForPrivateKey(string $uid):string {
		return $uid . '.private.key';
	}
}
