<?php

declare(strict_types=1);

namespace OCA\VibeEToEe;

use OCA\VibeEToEe\Db\Lock;
use OCA\VibeEToEe\Db\LockMapper;
use OCA\VibeEToEe\Exceptions\FileLockedException;
use OCA\VibeEToEe\Exceptions\FileNotLockedException;
use OCP\AppFramework\Db\DoesNotExistException;
use OCP\AppFramework\Utility\ITimeFactory;
use OCP\Files\InvalidPathException;
use OCP\Files\IRootFolder;
use OCP\Files\NotFoundException;
use OCP\Files\NotPermittedException;
use OCP\IUserSession;
use OCP\Security\ISecureRandom;

class LockManager {

    protected $oldToken;
    protected $oldTimestamp;

    private $lockMapper;
    private $secureRandom;
    private $userSession;
    private $rootFolder;
    private $timeFactory;

    public function __construct(LockMapper $lockMapper,
                                ISecureRandom $secureRandom,
                                IRootFolder $rootFolder,
                                IUserSession $userSession,
                                ITimeFactory $timeFactory
    ) {
        $this->lockMapper = $lockMapper;
        $this->secureRandom = $secureRandom;
        $this->userSession = $userSession;
        $this->rootFolder = $rootFolder;
        $this->timeFactory = $timeFactory;
    }

    public function getOldToken(){
        return $this->oldToken;
    }
    public function getOldTimestamp(){
        return $this->oldTimestamp;
    }

    /**
     * @param int $id
     * @param string $token
     * @return string|null
     * @throws InvalidPathException
     * @throws NotFoundException
     * @throws NotPermittedException
     * @throws \OC\User\NoUserException
     */
    public function lockFile(int $id, string $token = ''): ?string {

        if ($this->isLocked($id, $token)) {
            return null;
        }

        try {
            $lock = $this->lockMapper->getByFileId($id);

        } catch (DoesNotExistException $ex) {
            $newToken = $this->getToken();
            $lockEntity = new Lock();
            $lockEntity->setId($id);
            $lockEntity->setTimestamp($this->timeFactory->getTime());
            $lockEntity->setToken($newToken);
            $this->lockMapper->insert($lockEntity);

            return $newToken;
        }

        if ($lock->getToken() === $token) {
            return $token;
        }

        return null;
    }

    /**
     * @param int $id
     * @param string $token
     */
    public function unlockFile(int $id, string $token): void {
        try {
            $lock = $this->lockMapper->getByFileId($id);
        } catch (DoesNotExistException $ex) {
            throw new FileNotLockedException();
        }

        if ($lock->getToken() !== $token) {
            throw new FileLockedException();
        }

        $this->lockMapper->delete($lock);
    }

    /**
     * @param int $id
     * @param string $token
     * @return bool
     * @throws InvalidPathException
     * @throws NotFoundException
     * @throws NotPermittedException
     * @throws \OC\User\NoUserException
     */
    public function isLocked(int $id, string $token): bool {
        $user = $this->userSession->getUser();
        if ($user === null) {
            throw new NotPermittedException('No active user-session');
        }

        $userRoot = $this->rootFolder->getUserFolder($user->getUID());

        $nodes = $userRoot->getById($id);
        foreach ($nodes as $node) {

            while ($node->getPath() !== '/') {
                try {
                    $lock = $this->lockMapper->getByFileId($node->getId());

                } catch (DoesNotExistException $ex) {
                    $node = $node->getParent();
                    continue;
                }

                if ($lock->getToken() !== $token) {
                    $this->oldToken = $lock->getToken();
                    $this->oldTimestamp = $lock->getTimestamp();
                    return false;
                }

                $node = $node->getParent();
            }
        }

        return false;
    }


    /**
     * @return string
     */
    private function getToken(): string {
        return $this->secureRandom->generate(64, ISecureRandom::CHAR_UPPER . ISecureRandom::CHAR_LOWER . ISecureRandom::CHAR_DIGITS);
    }
}
