<?php

namespace OCA\VibeEToEe\Service;

use DateTime;
use Exception;
use OCP\IConfig;
use Psr\Log\LoggerInterface;
use OCP\IGroupManager;
use OCP\IUserManager;
use OCA\VibeEToEe\Contracts\IExportVibe;
use OCA\VibeEToEe\AppInfo\Application;
use OCA\VibeEToEe\Db\ExportMapper;
use OCA\VibeEToEe\Contracts\IRestApi;
use OCA\VibeEToEe\Db\UserAdminMapper;
use OCA\VibeEToEe\Db\GroupMapMapper;
use OCA\VibeEToEe\Db\NeUsersMapper;
use OCA\VibeEToEe\Service\UserAdminService as UAservice;

class ExportVibe implements IExportVibe
{
    const USER_BATCH_SIZE = 200;
    const WAMCUSER_BATCH_SIZE = 20;

    /** @var IRestApi */
	private $restApi;

    /** @var IGroupManager */
	private $groupManager;

    /** @var IUserManager */
    private $userManager;

    /** @var IConfig */
    protected $config;

    /** @var array */
    public $configVibe;

    /** @var LoggerInterface */
    private $logger;
    private $logInfo = [];

    /** @var array */
    private $users = [];

    /** @var array */
    private $idxSearch = [];

    /** @var ExportMapper */
    private $exportMapper;

    /** @var UserAdminMapper */
    private $userMapper;

    /** @var GroupMapMapper */
    private $groupMapper;

    /** @var NeUsersMapper */
    private $neUsersMapper;

    private $offset = 0;
    protected $uaservice;

    public function __construct(IConfig $config,
                                LoggerInterface $logger,
                                IGroupManager $groupManager,
                                IUserManager $userManager,
                                IRestApi $restApi,
                                ExportMapper $exportMapper,
                                UserAdminMapper $userMapper,
                                UAservice $uaservice,
                                NeUsersMapper $neUsersMapper,
                                GroupMapMapper $groupMapper ){
        $this->config = $config;
        $this->configVibe = Application::getConfigVibe();
        $this->logger = $logger;
        $this->groupManager = $groupManager;
        $this->userManager = $userManager;
        $this->restApi = $restApi;
        $this->exportMapper = $exportMapper;
        $this->userMapper = $userMapper;
        $this->groupMapper = $groupMapper;
        $this->neUsersMapper = $neUsersMapper;
        $this->uaservice = $uaservice;
    }

    /**
     * @param array $groups
     * @return $this
     * @throws Exception
     *
     * STEP 1
     */
    public function foundUsers(array $groups){
        $amountUsers = 0;
        try{
            foreach ($groups as $group) {
                $group = $this->groupManager->get($group);
                $amountUsers += $group->count();
            }
            $this->log('Found total users: ' . $amountUsers);

            $this->exportMapper->clearTable();

            return $this;

        } catch (Exception $e){
            throw $e;
        }

    }

    /**
     * @param int $offset
     * @return $this
     */
    public function setOffset(int $offset){
        $this->offset = $offset;
        return $this;
    }

    /**
     * @param string $group
     * @return array
     * @throws Exception
     *
     * STEP 2
     */
    public function exportFromGroup(string $group){
        try{
            $_group = $this->groupManager->get($group);
            $_users = $this->chkUsers($_group);
            if($_users) {
                $this->listExportToQueue($group);
            }
            return [
                'users' => $_users,
                'count' => $_group->count(),
                'perUser' => self::USER_BATCH_SIZE,
            ];

        } catch (Exception $e){
            throw $e;
        }
    }

    /**
     * @return array
     */
    public function getLog():array{
        return $this->logInfo;
    }

    /**
     * @throws Exception
     */
    public function executeByStep(){
        try{
            $users = (array) $this->getUsersForExport();
            if(count($users)){
                $this->usersToServerApi($users);
            }
            return true;

        } catch (Exception $e){
            throw $e;
        }
    }

    /**
     * @return array
     * @throws Exception
     */
    public function calculateProgress():array{
        try{
            return [
                'countAll' => $this->exportMapper->countUsers(),
                'countProgress' => $this->exportMapper->countUsers(true),
            ];

        } catch (Exception $e){
            throw $e;
        }

    }

    /**
     * @param array $userIds
     * @throws \OCP\AppFramework\OCS\OCSException
     * @throws \OCP\AppFramework\OCS\OCSNotFoundException
     * @throws \OC\OCS\Exception
     */
    public function usersListPrepare(array $userIds){
        $_users = [];
        foreach($userIds as $key => $userId){
            $_userAdmin = $this->userMapper->userFindById($userId);
            $_user = $this->uaservice->getInfoUser($_userAdmin->getUserId());

            $_users[] = [
                'identity' => $_user['id'],
                'email' => !empty($_user['email']) ? $_user['email'] : $key . 'none@test.local',
                'groups' => $_user['groups'],
            ];
        }
        $this->usersToServerApi($_users, false);
    }

    /**
     * @return array
     * @throws \OCP\AppFramework\Db\DoesNotExistException
     * @throws \OCP\AppFramework\Db\MultipleObjectsReturnedException
     * @throws \OCP\AppFramework\OCS\OCSException
     * @throws \OCP\AppFramework\OCS\OCSNotFoundException
     * @throws \OC\OCS\Exception
     */
    public function usersListPrepareByStep(){
        $_users = [];
        try{

            $usersNE = $this->neUsersMapper->paginate($this->offset, self::WAMCUSER_BATCH_SIZE);

            foreach($usersNE['data'] as $key => $userItem){
                $_user = $this->uaservice->getInfoUser($userItem->getUserId());

                $_users[] = [
                    'identity' => $_user['id'],
                    'email' => !empty($_user['email']) ? $_user['email'] : $key . 'none@test.local',
                    'groups' => $_user['groups'],
                ];
            }

            unset($usersNE['data']);
            $this->usersToServerApi($_users, false);

            return $usersNE;

        } catch (Exception $e){
            throw $e;
        }
    }

    /**
     * @param $users
     * @param bool $prepareUsers
     * @return bool
     * @throws \OC\OCS\Exception
     */
    protected function usersToServerApi($users, bool $prepareUsers = true){
        $_users = [];
        if($prepareUsers) {
            foreach ($users as $u) {
                $_users[] = [
                    'identity' => $u->getIdentity(),
                    'email' => $u->getEmail(),
                    'groups' => (array)json_decode($u->getData())
//                    'identity' => $u['identity'],
//                    'email' => $u['email'],
//                    'groups' => $u['email'],
                ];
//            array_push($data, $u->getData());
//            $this->log(sprintf('Preparing to send to the server: %s', $u->getIdentity()));
            }
        } else{
            $_users = $users;
        }

        $prepareData = json_encode([
            "encryptionMethod" => "none",
            "encryptedData" => [
                "users" => $_users
            ]
        ]);

        try{
            $result = (array) $this->sendToServer($prepareData);
            $this->parseResultServer($result, $_users);
            return true;

        } catch (Exception $e){
            $this->exportMapper->userProcessed(array_column($_users, 'identity'));
            throw $e;
        }
    }

    /**
     * @return \OCP\AppFramework\Db\Entity[]
     * @throws Exception
     */
    protected function getUsersForExport(){
        return $this->exportMapper->getUsers(false, self::WAMCUSER_BATCH_SIZE);
    }

    /**
     * @param $group
     * @return int
     * @throws Exception
     *
     * Creating a list of users to create a queue
     */
    protected function chkUsers($group){
        try{
            $users = $group->searchUsers('', self::USER_BATCH_SIZE, $this->offset * self::USER_BATCH_SIZE);
            $_countUsers = count($users);
            if(!$_countUsers)
                return $_countUsers;

            $userIds = [];
            foreach ($users as $user){
                $userIds[] = $user->getUID();
            }

            $linkIssue = $this->groupMapper->linkIssue($group->getGID(), $group->getGID());
            if($linkIssue){
                $userIds = $this->checkingUserExistence($userIds);
            }

            foreach ($users as $user){
                if(!in_array($user->getUID(), $userIds))
                    continue;

                array_push($this->users, [
                    "identity"  => $user->getUID(),
                    "email"     => $user->getEMailAddress(),
                    "groups"    => [$group->getGID()],
                ]);

            }
            return $_countUsers;

        } catch (Exception $e){
            throw $e;
        }
    }

    /**
     * @param $userId
     * @param $groupId
     * @return $this
     */
    protected function addingGroupToExistUser($userId, $groupId){
        $this->users[$this->idxSearch[$userId]]["groups"][] = $groupId;
        return $this;
    }

    /**
     * @param $userIds
     * @return \OCP\AppFramework\Db\Entity[]
     *
     * Validating users who already exist with the token.
     */
    protected function checkingUserExistence(array $userIds):array{
        $existsIds = $this->userMapper->getAllByIds($userIds, true, true);
        return array_diff($userIds, $existsIds);
    }

    /**
     * @param array $userIds
     * @return array
     *
     * Checking users for existence in the queue
     */
    protected function chkUsersQueue(array $userIds):array{
        $existsIds = $this->exportMapper->getAllUsersByIdentity($userIds);
        return array_diff($userIds, $existsIds);
    }

    /**
     * @param string $group
     * @return $this
     * @throws Exception
     */
    protected function listExportToQueue(string $group){
        try{
            $idsIssue = $this->chkUsersQueue(array_column($this->users, 'identity'));

            foreach ($this->users as $u){
                if(!in_array($u['identity'], $idsIssue)){
                    $_user = $this->exportMapper->userFind($u['identity']);
                    $_groups = (array) json_decode($_user->getData());
                    array_push($_groups, $group);
                    $_user->setData(json_encode($_groups));
                    $this->exportMapper->update($_user);

                } else {
                    $this->exportMapper->userCreate([
                        'identity'  => $u['identity'],
                        'email'     => $u['email'],
                        'data'      => json_encode($u['groups']),
                    ]);
                }
            }
            return $this;

        } catch (Exception $e){
            throw $e;
        }
    }

    /**
     * @param $data
     * @return mixed
     * @throws Exception
     */
    protected function sendToServer($data){
        try {
//            $this->log('Send to WAMC WEB API');

            $result = json_decode($this->restApi->setEndpoint('importUserList')
                ->setDataBody($data)
                ->run()
            );
//            $this->log('success WAMC WEB API');
            return $result;

        } catch (Exception $e){
//            $this->log('error WAMC WEB API');
            throw $e;
        }
    }

    /**
     * @param array $data
     * @param $_users
     */
    protected function parseResultServer(array $data, $_users)
    {
        foreach ($data['result'] as $u) {
            try {
                $this->exportMapper->userProcessed($u->identity);
//                $this->log($u->identity . ' success');

                $this->userMapper->userCreateToken($u->identity, $u->token);
//                $this->log($u->identity . ' add token');

                $key = array_search($u->identity, array_column($_users, 'identity'));
                $this->linkGroups($_users[$key]);

            } catch (Exception $e) {
                $this->log($u->identity . ' error. ' . $e->getMessage());
                continue;
            }
        }
    }

    /**
     * @param $user
     * @return bool
     * @throws \OCP\AppFramework\Db\DoesNotExistException
     * @throws \OCP\AppFramework\Db\MultipleObjectsReturnedException
     */
    protected function linkGroups($user)
    {
        foreach ($user['groups'] as $g) {
            try {
                if($this->groupMapper->createLink($g)) {
                    $this->log(sprintf('Group "%s" linked', $g));
                }

            } catch (Exception $e) {
                $this->log($g . ' not linked. ' . $e->getMessage());
                continue;
            }
        }
        return true;
    }

    /**
     * @param string $msg
     * @return bool
     */
    protected function log(string $msg){
        $date = new DateTime('now');
        $this->logInfo[] = [
            'date' => $date->format('d.m.Y h:i:s'),
            'msg' => $msg,
        ];
        return true;
    }
}
